﻿using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using static JSONStorableBool;
using static JSONStorableFloat;
using static JSONStorableString;
using static MiscreatedFinger.PoseJoints;
using sc = SuperController;

namespace MiscreatedFinger
{
    public class FingerReAnimator : MVRScript
    {
        const string pluginName = "FingerReAnimator";
        const string pluginVersion = "V1.3.0";
        const string pluginAuthor = "Miscreated Reality";
        const string pluginLink = "www.MiscreatedReality.com";
        const string license = "PC-Paid Content, Cannot distribute, remix, tweak, or build upon work";

        public static Dictionary<string, ConfigurableJoint> fingerJoints = new Dictionary<string, ConfigurableJoint>();
        public static Dictionary<string, ConfigurableJoint> thumbJoints = new Dictionary<string, ConfigurableJoint>();

        public JSONStorableFloat timeline = new JSONStorableFloat("Timeline [s]", 0f, 0f, 0f);
        public JSONStorableFloat playbackSpeed = new JSONStorableFloat("Playback speed", 0f, -5f, 5f);
        public JSONStorableFloat playbackTimeOffset = new JSONStorableFloat("Time Offset", 0f, -60f, 60f);
        private JSONStorableString loadedAnimation = new JSONStorableString("Loaded Animation", "");
        private JSONStorableBool autoPlay = new JSONStorableBool("Auto play", true);
        private JSONStorableBool useTimeline = new JSONStorableBool("Use Timeline", true);

        public List<PoseJoints> storedJointRotations = new List<PoseJoints>();

        public int currentPoseIndex = -1;
        public PoseJoints currentPose
        {
            get { return currentPoseIndex < 0 ? null : storedJointRotations[currentPoseIndex]; }
        }

        string defaultDir = @"Saves/PluginData/miscreated/ReAnimations";
        public static List<string> FingerParts = new List<string>() { "lRing1", "lRing2", "lRing3", "lPinky1", "lPinky2", "lPinky3", "lIndex1", "lIndex2", "lIndex3", "lMid1", "lMid2", "lMid3", "lThumb2", "lThumb3", "rRing1", "rRing2", "rRing3", "rPinky1", "rPinky2", "rPinky3", "rIndex1", "rIndex2", "rIndex3", "rMid1", "rMid2", "rMid3", "rThumb2", "rThumb3" };
        public static List<string> ThumbParts = new List<string>() { "lThumb1", "rThumb1" };
        private string fileExtension = "";
        private JSONStorableFloat force = new JSONStorableFloat("Max Force", 1, 0.1f, 5f);
        private JSONStorableFloat damper = new JSONStorableFloat("Damper", 1f, 0.1f, 5f);
        private MotionAnimationMaster motion;
        private Func<float> getTime;
        private MVRScript timelinePlugin = null;

        public override void Init()
        {
            try
            {
                if (containingAtom.type != "Person")
                {
                    throw new Exception("For person atoms only!");
                }
                CreateUI();
                fingerJoints = new Dictionary<string, ConfigurableJoint>();

                AddJoints("lHand", FingerParts, fingerJoints);
                AddJoints("lHand", ThumbParts, thumbJoints);
                AddJoints("rHand", FingerParts, fingerJoints);
                AddJoints("rHand", ThumbParts, thumbJoints);

                loadedAnimation.setJSONCallbackFunction += (JSONStorableString s) =>
                {
                    Load(s.val);
                };

                //RegisterString(loadedAnimation);
                //loadedAnimation.storeType = JSONStorableParam.StoreType.Full;

                motion = sc.singleton.motionAnimationMaster;
                getTime = () => motion.GetCurrentTimeCounter();

                if(useTimeline.val)
                {
                    StartCoroutine(TimelineSearch(containingAtom));
                }

                sc.LogMessage($"Finger ReAnimator Init OK");
            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(Init)} error: {x.Message}");
            }
        }

        private void ReconfigureFingerJoints(Dictionary<string, ConfigurableJoint> joints, float f, float d)
        {
            foreach (var jk in fingerJoints)
            {
                jk.Value.angularXDrive = new JointDrive { maximumForce = f * 2f, positionSpring = f * 5f, positionDamper = d * 0.1f };
            }
        }
        private void ReconfigureThumbJoints(Dictionary<string, ConfigurableJoint> joints, float f, float d)
        {
            foreach (var jk in thumbJoints)
            {
                jk.Value.slerpDrive = new JointDrive { maximumForce = f * 50f, positionSpring = f * 40f, positionDamper = d * 2f };
            }
        }

        private void CreateUI()
        {
            try
            {
                Button("Load", true, onLoad);
                Text(loadedAnimation, true, (JSONStorableString s) => { Load(s.val); });
                Toggle(autoPlay);
                Toggle(useTimeline, false, (JSONStorableBool b) => { if (b.val) StartCoroutine(TimelineSearch(containingAtom)); else { getTime = () => motion.GetCurrentTimeCounter(); timelinePlugin = null; } });
                Slider(force, true, AdjustFingers);
                Slider(damper, true, AdjustFingers);
                Slider(playbackSpeed);
                Slider(timeline, false, (JSONStorableFloat f) => { TimeLineValueUpdated(f.val); });
                Slider(playbackTimeOffset);
            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(CreateUI)} error: {x.Message}");
            }
        }

        private void AdjustFingers(JSONStorableFloat jf)
        {
            ReconfigureFingerJoints(fingerJoints, force.val, damper.val);
            ReconfigureThumbJoints(thumbJoints, force.val, damper.val);
        }

        private void AddJoints(string storableId, List<string> filter, Dictionary<string, ConfigurableJoint> addJoints)
        {
            try
            {
                var joints = containingAtom.GetStorableByID(storableId).GetComponentsInChildren<ConfigurableJoint>();
                foreach (var item in joints)
                {
                    if (filter.Contains(item.name))
                    {
                        addJoints.Add(item.name, item);
                    }
                }

            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(AddJoints)} error: {x.Message}");
            }
        }

        private void Animate()
        {
            try
            {
                if (storedJointRotations.Count == 0) return;

                float time = timeline.val;
                float delta = playbackSpeed.val * Time.fixedDeltaTime;
                bool overflow = time + delta > timeline.max || time + delta < 0;

                time += (playbackSpeed.val < 0 ? timeline.max : 0) + delta;

                if (overflow)
                {
                    playbackSpeed.val = 0f;
                }

                timeline.valNoCallback = TimeLineValueUpdated(time);
            }
            catch (Exception x)
            {
                sc.LogMessage($"{nameof(FingerReAnimator)} error: {x.Message}");
            }
        }
        private float TimeLineValueUpdated(float time)
        {
            if (storedJointRotations.Count == 0) return time;

            if (time > timeline.max)
            {
                time %= timeline.max;
            }

            if (currentPose == null || time < currentPose.Time || time > currentPose.Time + 0.03f)
            {
                int currentIndex = storedJointRotations.FindIndex(x => x.Time > time);

                if (currentIndex != currentPoseIndex)
                {
                    currentPoseIndex = currentIndex;
                    UpdateJointRotations();
                }
            }

            return time;
        }

        public void UpdateJointRotations()
        {
            try
            {
                if (currentPose != null)
                {
                    foreach (var jointTargetDrive in currentPose.Joints)
                    {
                        jointTargetDrive.Value.Joint.targetRotation = jointTargetDrive.Value.Target;
                    }
                }
            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(UpdateJointRotations)} error: {x.Message}");
            }
        }

        public void Load(string filename)
        {
            try
            {
                playbackSpeed.val = 0f;
                loadedAnimation.valNoCallback = filename;
                sc.LogMessage($"Load Pose {filename}");
                if (string.IsNullOrEmpty(filename)) return;

                filename = Extensions.GetRealFilename(this, filename);

                var jc = sc.singleton.LoadJSON(filename);

                storedJointRotations.Clear();

                foreach (JSONClass pose in jc["poses"].AsArray)
                {
                    storedJointRotations.Add(new PoseJoints(pose, fingerJoints));
                }

                sc.LogMessage($"Loaded {filename} - {storedJointRotations.Count} samples.");

                timeline.max = storedJointRotations.Max(x => x.Time);
                CleanAnimation();
            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(Load)} error: {x.Message}");
            }
        }
        private void CleanAnimation()
        {
            if (storedJointRotations.Count == 0) return;

            var allJoints = storedJointRotations.First().Joints.Select(x => x.Key).ToList();
            var missingJoints = allJoints.Except(fingerJoints.Keys);

            foreach (var pose in storedJointRotations)
            {
                foreach (var missing in missingJoints)
                {
                    pose.Joints.Remove(missing);
                }
            }
        }
        private void onLoad()
        {
            try
            {
                sc.singleton.GetMediaPathDialog(Load, fileExtension, defaultDir, false, true, false);

                uFileBrowser.FileBrowser browser = sc.singleton.mediaFileBrowserUI;
                browser.SetTitle("Select animation file to load");

                browser.fileEntryField.text = string.IsNullOrEmpty(loadedAnimation.val) ? string.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), fileExtension) : FileManagerSecure.GetFileName(loadedAnimation.val);
                browser.ActivateFileNameField();
            }
            catch (Exception x)
            {
                sc.LogError($"{nameof(FingerReAnimator)}.{nameof(onLoad)} error: {x.Message}");
            }

        }

        void FixedUpdate()
        {
            if (autoPlay.val)
            {
                timeline.val = getTime() + playbackTimeOffset.val;
            }
            if (playbackSpeed.val != 0 && !sc.singleton.freezeAnimation)
            {
                Animate();
            }
        }

        void OnDestroy()
        {
            ReconfigureFingerJoints(fingerJoints, 1f, 1f);
            ReconfigureThumbJoints(thumbJoints, 1f, 1f);
        }

        protected UIDynamicToggle Toggle(JSONStorableBool jsb, bool rightSide = false, SetJSONBoolCallback callback = null)
        {
            var toggle = this.CreateToggle(jsb, rightSide);
            jsb.setJSONCallbackFunction = callback;
            RegisterBool(jsb);
            jsb.storeType = JSONStorableParam.StoreType.Full;
            return toggle;
        }
        protected UIDynamicSlider Slider(JSONStorableFloat jsf, bool rightSide = false, SetJSONFloatCallback callback = null)
        {
            var slider = this.CreateSlider(jsf, rightSide);
            if (callback != null)
            {
                jsf.setJSONCallbackFunction = callback;
            }
            RegisterFloat(jsf);
            jsf.storeType = JSONStorableParam.StoreType.Full;
            return slider;
        }
        protected UIDynamicTextField Text(JSONStorableString js, bool rightSide = false, SetJSONStringCallback callback = null)
        {
            var slider = this.CreateTextField(js, rightSide);
            if (callback != null)
            {
                js.setJSONCallbackFunction = callback;
            }
            RegisterString(js);
            js.storeType = JSONStorableParam.StoreType.Full;
            return slider;
        }
        protected UIDynamicButton Button(string label, bool rightSide, UnityEngine.Events.UnityAction action)
        {
            var btn = CreateButton(label, rightSide);
            btn.button.onClick.AddListener(action);
            return btn;
        }
        private IEnumerator TimelineSearch(Atom atom, float interval = 1, int max = 5)
        {
            sc.singleton.Message($"Start searching for timeline plugin.");
            while (atom && !timelinePlugin && max-- > 0)
            {
                yield return new WaitForSeconds(interval);
                timelinePlugin = atom.gameObject.GetComponentsInChildren<MVRScript>().FirstOrDefault(x => x.name.Contains("Timeline"));
                if (timelinePlugin)
                {
                    var time = timelinePlugin.GetFloatJSONParam("Scrubber");
                    if (time != null)
                    {
                        sc.singleton.Message("found time");
                        getTime = () => time.val;
                    }
                    else
                    {
                        sc.singleton.Message("didn't find time");

                    }
                    break;
                }
            }
            sc.singleton.Message($"{nameof(FingerReAnimator)}, found timeline plugin.");
        }
    }

    public class PoseJoints
    {
        public string Name;
        public Dictionary<string, JointTargetDrive> Joints = new Dictionary<string, JointTargetDrive>();

        public float Time = 1f;

        public JSONClass ToJson(bool useMuscles = true)
        {
            var joints = new JSONArray();

            foreach (var joint in Joints)
            {
                joints.Add(Name, joint.Value.JSONClass(useMuscles));
            }

            var pose = new JSONClass
            {
                ["name"] = Name,
                ["joints"] = joints,
            };
            pose["time"].AsFloat = Time;

            return pose;
        }
        public PoseJoints()
        {

        }
        public class JointTargetDrive
        {
            public JointTargetDrive()
            {
            }
            public string Name;
            public Quaternion Target;
            public ConfigurableJoint Joint = null;

            public JointTargetDrive Clone()
            {
                var clone = new JointTargetDrive
                {
                    Name = this.Name,
                    Target = this.Target,
                };
                return clone;
            }
        }
        public PoseJoints(JSONClass poseJoints, Dictionary<string, ConfigurableJoint> joints = null)
        {
            Joints.Clear();
            Name = poseJoints["name"];
            Time = poseJoints["time"].AsFloat;

            foreach (JSONClass node in poseJoints["joints"].AsArray)
            {
                if (string.IsNullOrEmpty(node["name"])) continue;

                string jointKey = node["name"];
                var j = new JointTargetDrive
                {
                    Target = Quaternion.Euler(node["x"].AsFloat, node["y"].AsFloat, node["z"].AsFloat),
                    Name = jointKey,
                };

                if (joints != null && joints.ContainsKey(jointKey))
                {
                    j.Joint = joints[jointKey];
                }

                if (!Joints.ContainsKey(node["name"]))
                {
                    Joints.Add(node["name"], j);
                }
            }
        }
    }
    public static class Extensions
    {
        public static float Degree180(this float val)
        {
            if (Mathf.Approximately(val, 0f)) return 0f;
            if (Mathf.Approximately(val, 360f)) return 0f;

            if (val > 180f)
            {
                return val - 360f;
            }
            if (val < -180f)
            {
                return val + 360f;
            }
            return val;
        }
        public static JSONClass JSONClass(this JointTargetDrive targetdrive, bool useMuscles = true)
        {
            var node = new JSONClass
            {
                ["name"] = targetdrive.Name
            };

            node["x"].AsFloat = targetdrive.Target.eulerAngles.x.Degree180();
            node["y"].AsFloat = targetdrive.Target.eulerAngles.y.Degree180();
            node["z"].AsFloat = targetdrive.Target.eulerAngles.z.Degree180();
            return node;
        }
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }
        public static string GetPackagePath(MVRScript self)
        {
            string filename = GetPluginPath(self);
            int idx = filename.IndexOf(":/");
            if (idx >= 0)
                return filename.Substring(0, idx + 2);
            else
                return string.Empty;
        }
        public static bool IsInPackage(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.IndexOf(":/") >= 0;
        }
        public static string GetRealFilename(MVRScript self, string filename)
        {
            if (IsInPackage(self))
            {
                var package = GetPackagePath(self);
                sc.singleton.Message($"Package = {package}");
                return $"{package}{filename}";
            }
            return filename;
        }
    }
}
